from selenium.webdriver.remote.webelement import WebElement

from change_analyzer.spaces.actions.app_action import AppAction


class InputAppAction(AppAction):
    def __init__(self, el: WebElement, value: str) -> None:
        super(InputAppAction, self).__init__(el)
        self.value = value

    def perform(self) -> None:
        self.el.clear()
        self.el.send_keys(self.value)

    def __str__(self) -> str:
        return f"fill input {self.el.get_attribute('name')} with {self.value}"



from io import BytesIO
import time
from typing import Callable, Tuple, Dict, Union

import numpy as np
from PIL import Image, ImageGrab
from pywinauto import Desktop, Application
from selenium.webdriver.remote.webdriver import WebDriver

from change_analyzer.envs.web_driver_env import WebDriverEnv
from change_analyzer.spaces.actions.app_action import AppAction


class AppEnv(WebDriverEnv):
    def __init__(self, reset_app: Callable[[], WebDriver]) -> None:
        super(AppEnv, self).__init__(reset_app)
        self._initial_windows_without_sut = []

    def reset(self) -> Union[Dict, np.ndarray]:
        self._initial_windows_without_sut = self._get_current_open_windows()
        self._driver = self._reset_app()
        self._update_spaces()
        return self._observe()

    def step(self, action: AppAction) -> Tuple[Dict, float, bool, WebDriver]:
        # reward options:
        # - [CURRENT]reward if no exception was rised during the action
        # - len(self.action_space.actions - prev_action_space.actions) - we would like to find more new actions
        reward = 0
        try:
            self._logger.info(f"Perform action {str(action)}")
            action.perform()
            reward = 1
        except Exception as e:
            self._logger.error(e)
        finally:
            self._update_spaces()
            obs = self._observe()
            done = len(self.action_space.actions) == 0 or self._we_have_new_windows()
            info = self._driver
            return obs, reward, done, info

    def close(self) -> None:
        try:
            self._driver.close()
            self._driver.quit()
        except Exception as e:
            self._logger.info(
                "Due to an exception the driver couldn't be closed or quit"
            )
            self._logger.info(e)
        finally:
            # As one last thing to do when closing the env is to clean the opened windows
            time.sleep(2)
            current_open_windows = self._get_current_open_windows()

            for w in current_open_windows:
                if "F-Secure" in w or "Software updates" in w:
                    self._kill_app_based_on_window_title(window_title=w)

            # while not current_open_windows == initial_windows_without_sut:
            #     for w in current_open_windows:
            #         if w not in initial_windows_without_sut:
            #             self.kill_app_based_on_window_title(window_title=w)
            #     current_open_windows = self.get_current_open_windows()

    @staticmethod
    def _get_current_open_windows() -> list:
        """Get a list of all open windows"""

        windows = Desktop(backend="uia").windows()
        open_windows = [w.window_text() for w in windows]

        return open_windows

    def _kill_app_based_on_window_title(self, window_title: str):
        """Kill the app associated with the window title"""
        try:
            window = Desktop(backend="uia").windows(
                title=window_title, control_type="Window"
            )[0]
        except Exception as e:
            self._logger.info(
                f"Exception encountered when trying to find the window based on title {window_title}"
            )
            self._logger.error(e)
        # Get the window's handle
        try:
            window_handle = window.handle
            app = Application().connect(handle=window_handle)
            app.kill()
        except Exception as e:
            self._logger.info(
                f"Exception encountered when trying to kill the app based on title {window_title}"
            )
            self._logger.error(e)

    def _we_have_new_windows(self) -> bool:
        """Check if we have new windows, comparing the length of initial windows and current windows
        Initial windows do not contain the SUT window"""
        return (
                len(self._get_current_open_windows())
                > len(self._initial_windows_without_sut) + 1
        )

    def _get_screenshot(self) -> Image:
        try:
            stream = BytesIO(self._driver.get_screenshot_as_png())
            pic = Image.open(stream).convert("RGB")
            stream.close()
            return pic
        except Exception as e:
            self._logger.error(e)
            self._logger.info(f"Cannot get app screenshot. Taking full screen image")
            return ImageGrab.grab()



from PIL import Image
import os

from change_analyzer.utils import image_pad_resize_to


def test_image_pad_resize_to_wide():
    size = (800, 400)
    cwd = os.getcwd()
    image_from_disk = Image.open(f"{cwd}/change_analyzer/tests/400x400.png")
    im = image_pad_resize_to(image_from_disk, size)
    assert im.size == size


def test_image_pad_resize_to_square():
    size = (800, 800)
    cwd = os.getcwd()
    image_from_disk = Image.open(f"{cwd}/change_analyzer/tests/400x400.png")
    im = image_pad_resize_to(image_from_disk, size)
    assert im.size == size


def test_image_pad_resize_to_tall():
    size = (400, 800)
    cwd = os.getcwd()
    image_from_disk = Image.open(f"{cwd}/change_analyzer/tests/400x400.png")
    im = image_pad_resize_to(image_from_disk, size)
    assert im.size == size



import gym as gym
from gym.wrappers import Monitor

from change_analyzer.wrappers.transparent_wrapper_mixin import TransparentWrapperMixin


class EnhancedMonitor(Monitor, TransparentWrapperMixin):
    def __init__(
        self,
        env: gym.Env,
        directory: str,
        video_callable=None,
        force=False,
        resume=False,
        write_upon_reset=False,
        uid=None,
        mode=None,
    ) -> None:
        super(EnhancedMonitor, self).__init__(
            env, directory, video_callable, force, resume, write_upon_reset, uid, mode
        )



from selenium.webdriver.remote.webelement import WebElement


class AppAction:
    def __init__(self, el: WebElement) -> None:
        # TODO replace el with image. GUI testing is an visual task! can do `el.screenshot()` here
        self.el = el

    def perform(self) -> None:
        pass

    # def _get_xpath(self, el, current: str = "") -> Union[str, None]:
    #     el_tag = el.tag_name
    #     try:
    #         parent = el.find_element_by_xpath("..")
    #     except:
    #         return f"/{el_tag}[1]{current}"
    #
    #     count = 0
    #     children = parent.find_elements_by_xpath("*")
    #     for child in children:
    #         if el_tag == child.tag_name:
    #             count = count + 1
    #         if el.id == child.id:
    #             return self._get_xpath(parent, f"/{el_tag}[{count}]{current}")
    #     return None



import os
from os.path import isfile
from typing import Dict, Tuple
import pandas as pd
import time
import json

import gym as gym
from gym import Wrapper
from selenium.webdriver.remote.webdriver import WebDriver
from lxml import html, etree

from change_analyzer.spaces.actions.app_action import AppAction
from change_analyzer.wrappers.transparent_wrapper_mixin import TransparentWrapperMixin


class SequenceRecorder(Wrapper, TransparentWrapperMixin):
    COL_TIMESTAMP = 'Timestamp'
    COL_SEQUENCE_ID = 'SequenceId'
    COL_PAGE_SOURCE_BEFORE = 'PageSourceBefore'
    COL_PAGE_SOURCE_AFTER = 'PageSourceAfter'
    COL_ACTION_TO_PERFORM = 'ActionToPerform'
    COL_ACTION_IMAGE_BEFORE = 'ActionImageBefore'
    COL_ACTION_IMAGE_AFTER = 'ActionImageAfter'

    def __init__(self, env: gym.Env, directory: str, sequence_id: str) -> None:
        super(SequenceRecorder, self).__init__(env)
        self._sequence_id = sequence_id
        os.makedirs(directory, exist_ok=True)
        self._csv_file = f"{directory}/{sequence_id}.csv"

    def step(self, action: AppAction) -> Tuple[Dict, float, bool, WebDriver]:
        current_action = str(action)
        page_source_before = self.get_enriched_page_source(info=action.el.parent)
        page_source_after = ""
        image_before = []
        image_after = []

        try:
            image_before = self.env.render("rgb_array").tolist()
            obs, reward, done, info = self.env.step(action)
            image_after = self.env.render("rgb_array").tolist()
            page_source_after = self.get_enriched_page_source(info)
        finally:
            self._save_dataframe_to_csv(pd.DataFrame({
                self.COL_TIMESTAMP: [int(time.time())],
                self.COL_SEQUENCE_ID: [self._sequence_id],
                self.COL_PAGE_SOURCE_BEFORE: [page_source_before],
                self.COL_PAGE_SOURCE_AFTER: [page_source_after],
                self.COL_ACTION_TO_PERFORM: [current_action],
                self.COL_ACTION_IMAGE_BEFORE: [json.dumps(image_before)],
                self.COL_ACTION_IMAGE_AFTER: [json.dumps(image_after)],
            }))
        return obs, reward, done, info

    @staticmethod
    def get_enriched_page_source(info: WebDriver) -> str:
        """Enrich a given page source with additional info from the WebDriver
        All page elements will have the following additional info:
            - Text if available
            - x, y coordinates
            - h, w dimensions
        """
        # Get elements' by xpath from given page source
        root = html.fromstring(info.page_source)
        tree = root.getroottree()
        all_elements_by_xpath = root.xpath('//*')
        for element in all_elements_by_xpath:
            element_xpath = tree.getpath(element)
            xpath_to_find = "./"
            if 'head' in element_xpath:
                xpath_to_find = f"./{element_xpath.split('/html/head')[1]}"

            if 'body' in element_xpath:
                xpath_to_find = f"./{element_xpath.split('/html/body')[1]}"

            if xpath_to_find != "./":
                driver_el = info.find_element_by_xpath(element_xpath)
                el = root.find(xpath_to_find)
                el.set('x', f"{driver_el.location['x']}")
                el.set('y', f"{driver_el.location['y']}")
                el.set('height', f"{driver_el.size['height']}")
                el.set('width', f"{driver_el.size['width']}")

        # Return the enriched page source decoded from bytes
        return etree.tostring(root).decode("utf-8")

    def close(self) -> None:
        self.env.close()

    def _save_dataframe_to_csv(self, df: pd.DataFrame) -> None:
        if isfile(self._csv_file):  # it exists so append without writing the header
            df.to_csv(self._csv_file, index=False, mode='a', header=False)
        else:
            df.to_csv(self._csv_file, index=False)



import glob
import os
from typing import List

import gym

from change_analyzer.agents.agent import Agent
from change_analyzer.spaces.actions.app_action import AppAction
import pandas as pd
import logging


class ReplayAgent(Agent):
    def __init__(self, env: gym.Env, csv_folder: str) -> None:
        super(ReplayAgent, self).__init__(env)
        self._logger = logging.getLogger(__name__)
        self.csv_file = glob.glob(os.path.join(os.getcwd(), csv_folder, "*.csv"))[0]
        self.replay_actions = self._update_actions_from_csv()

    def run(self) -> None:
        try:
            for replay_action in self.replay_actions:
                action = self._find_action(replay_action)
                print(f"Perform action {str(action)}")
                _, _, done, _ = self.env.step(action)
                if done:
                    break
        except Exception as e:
            self._logger.info("Running failed due to an exception")
            self._logger.info(e)

    def _find_action(self, replay_action_to_find: str) -> AppAction:
        for action in self.env.action_space.actions:
            if str(action) == replay_action_to_find:
                return action

        raise Exception(f"Action {replay_action_to_find} was not found")

    def _update_actions_from_csv(self) -> List[str]:
        """Extract actions from given CSV file"""
        df = pd.read_csv(self.csv_file)

        return list(df['ActionToPerform'])



__version__ = "0.10.0"

import sys

from gym.envs.registration import register


register(
    id="web-v0",
    entry_point="change_analyzer.envs:WebDriverEnv",
)
if sys.platform == 'win32':
    register(
        id="app-v0",
        entry_point="change_analyzer.envs:AppEnv",
    )



import itertools
import logging
import random
from typing import List

from faker import Faker
from gym import Space
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement

from change_analyzer.spaces.actions.app_action import AppAction
from change_analyzer.spaces.actions.click_app_action import ClickAppAction


class DiscreteAppActionSpace(Space):
    ACTIONABLE_SELECTORS = {
        "web": {
            "buttons": {"button", "a", "select", "option", "input"},
            "inputs": {"input", "textarea"},
        },
        "win": {"buttons": {"Button", "MenuItem"}, "inputs": set()},
        "mobile": {"buttons": set(), "inputs": set()},
    }
    OVERLAPING_PERCENTAGE = 0.7
    SEED = 123
    NUMBER_OF_TEXT_SAMPLES = 3
    LOCALES = {"en_US", "fi_FI", "ja_JP"}

    def __init__(
        self,
        driver: WebDriver,
    ) -> None:
        super(DiscreteAppActionSpace, self).__init__((), AppAction)
        self._logger = logging.getLogger(__name__)
        self._fake = Faker(self.LOCALES)
        self._fake.seed_instance(self.SEED)
        self.actions = set()
        self.driver = driver
        self.actionable_selectors = set()
        self.update_actionable_selectors()
        self.actionable_elements = set()
        self.update_actionable_elements()
        self.update_current_actions()

    def update_actionable_selectors(self):
        self.actionable_selectors = self.ACTIONABLE_SELECTORS["win"]
        try:
            if self.driver.find_element_by_xpath("/*").tag_name == "html":
                self.actionable_selectors = self.ACTIONABLE_SELECTORS["web"]
        except Exception as e:
            self._logger.info(
                "Actionable selectors couldn't be updated due to an exception"
            )
            self._logger.info(e)

    def update_actionable_elements(self):
        try:
            self.actionable_elements = {
                el
                for selector in set.union(*self.actionable_selectors.values())
                for el in self.driver.find_elements_by_xpath(f"//*/{selector}")
                if self._is_el_clickable(el)
            }
            for el0, el1 in itertools.combinations(self.actionable_elements, 2):
                if self._overlap_percentage(el0, el1) > self.OVERLAPING_PERCENTAGE:
                    self.actionable_elements.remove(el1)
                if self._overlap_percentage(el1, el0) > self.OVERLAPING_PERCENTAGE:
                    self.actionable_elements.remove(el0)
        except Exception as e:
            self._logger.info(
                "Actionable elements couldn't be updated due to an exception"
            )
            self._logger.info(e)

    def update_current_actions(self):
        for el in self.actionable_elements:
            # if el.tag_name in actionable_selectors["buttons"]:
            self.actions.add(ClickAppAction(el))
            # if el.tag_name in actionable_selectors["inputs"]:
            #     # TODO how do we handle checkboxes, dropdowns, radios somehow?
            #     self.actions.update(
            #         {InputAppAction(el, value) for value in self._generate_values(el)}
            #     )
        self._logger.info(f"Found {len(self.actions)} actions on the screen")

    def sample(self) -> AppAction:
        return random.choice(list(self.actions))

    def contains(self, x: str) -> bool:
        # TODO implement __eq__() in AppAction
        return x in self.actions

    def __repr__(self) -> str:
        return f"DiscreteAppAction({self.actions})"

    def __eq__(self, other: Space) -> bool:
        # TODO implement __eq__() in AppAction
        return (
            isinstance(other, DiscreteAppActionSpace) and self.actions == other.actions
        )

    def _overlap_percentage(self, el0, el1) -> float:
        # TODO check that the elements are not overlaping
        # self._logger.info(f"checking rects: {el0.rect}, {el1.rect}")
        # https://stackoverflow.com/a/42874377 or more compact: https://stackoverflow.com/a/57247833
        return 0

    def _is_el_clickable(self, el: WebElement) -> bool:
        try:
            return (
                (el.rect["width"] != 0 or el.rect["height"] != 0)
                and el.is_enabled()
                and el.is_displayed()
            )
        except Exception as e:
            self._logger.info(
                f"Checking if element is clickable failed due to an exception due to an exception"
            )
            self._logger.info(e)
            return False

    def _generate_values(self, el: WebElement) -> List[str]:
        # TODO generate at least one valid value basing on input.get_attribute("name/type/min/max") and several invalid
        providers = {"address", "company", "email", "name", "phone_number", "ssn"}
        return [
            getattr(self._fake, provider)()
            for provider in providers
            for i in range(self.NUMBER_OF_TEXT_SAMPLES * len(self.LOCALES))
        ]



from selenium.webdriver.remote.webelement import WebElement

from change_analyzer.spaces.actions.app_action import AppAction


class ClickAppAction(AppAction):
    def __init__(self, el: WebElement) -> None:
        super(ClickAppAction, self).__init__(el)

    def perform(self) -> None:
        self.el.click()

    def __str__(self) -> str:
        if self.el.text:
            return f"click on {self.el.text}"
        elif self.el.get_attribute("HelpText"):
            return f"click on {self.el.get_attribute('HelpText')}"
        elif self.el.tag_name:
            return f"click on {self.el.tag_name}"
        else:
            return f"click on element from location ({self.el.location})"



import uuid
from configparser import ConfigParser
from datetime import datetime
import argparse

import gym
from selenium import webdriver
from selenium.webdriver import DesiredCapabilities
from selenium.webdriver.remote.webdriver import WebDriver

from change_analyzer.agents.random_agent import RandomAgent
from change_analyzer.agents.replay_agent import ReplayAgent
from change_analyzer.wrappers.enhanced_monitor import EnhancedMonitor
from change_analyzer.wrappers.sequence_recorder import SequenceRecorder


CONFIG = ConfigParser()


def reset() -> WebDriver:
    capabilities = DesiredCapabilities.CHROME
    if CONFIG["driver"]["platform"] == "win":
        capabilities = {
            "app": CONFIG["driver"]["app"],
            "platformName": "Windows",
            "deviceName": "WindowsPC",
        }
        driver = webdriver.Remote(CONFIG["driver"]["command_executor"], capabilities)

    if CONFIG["driver"]["platform"] == "web":
        web_options = webdriver.ChromeOptions()
        web_options.add_argument("--disable-dev-shm-usage")
        web_options.add_argument("--start-maximized")
        web_options.add_argument("--headless")
        driver = webdriver.Remote(CONFIG["driver"]["command_executor"], capabilities, options=web_options)
        driver.get(CONFIG["driver"]["url"])

    # if self.driver is not None:
    # - get another available VM and connect to it
    # - kill/reset current one

    return driver


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        help="path to config file",
        required=True,
    )
    parser.add_argument(
        "--csv_folder",
        help="path to the folder within recordings which has the targeted csv file",
        required=False,
    )
    args = parser.parse_args()
    CONFIG.read(args.config)

    env = gym.make(
        "app-v0" if CONFIG["driver"]["platform"] == "win" else "web-v0",
        reset_app=reset,
    )
    # For some reason registration sometimes doesn't work and line above can fail, direct class creation as in line
    # below could be used. See https://github.com/openai/gym/blob/master/docs/creating-environments.md for more info
    # about registration.
    # env = AppEnv(reset, {"Help"})
    sequence_id = uuid.uuid1()
    report_dir = "recordings/{}".format(datetime.now().strftime("%Y_%m_%d-%H_%M_%S"))
    # env = EnhancedMonitor(env, report_dir)
    env = SequenceRecorder(env, report_dir, sequence_id)
    try:
        env.reset()
        if args.csv_folder:
            ReplayAgent(env, args.csv_folder).run()
        else:
            RandomAgent(env, 5).run()
    finally:
        env.close()


if __name__ == "__main__":
    main()



import gym

from change_analyzer.agents.agent import Agent


class StrategyAgent(Agent):
    def __init__(self, env: gym.Env) -> None:
        super(StrategyAgent, self).__init__(env)

    def run(self) -> None:
        # 1. find all the screens and identify forms
        # 2. record it as a test
        # 3. reset
        # for form in forms:
        #     1. navigate to the form
        #     2. test form with invalid inputs
        #     3. record it as a test
        #     4. test form with valid inputs
        #     5. search your input on the "very next page"/"somewhere around" and make an assertion
        #     6. record it as a test
        #     7. reset
        pass



import base64
import logging
from io import BytesIO
from typing import Any, Callable, Tuple, Dict, Union

import gym
import numpy as np
from PIL import Image
from gym import spaces
from selenium.webdriver.remote.webdriver import WebDriver

from change_analyzer.spaces.actions.app_action import AppAction
from change_analyzer.spaces.discrete_app_action_space import DiscreteAppActionSpace
from change_analyzer.utils import image_pad_resize_to


class WebDriverEnv(gym.Env):
    metadata = {
        "render.modes": ["human", "base64", "rgb_array"],
        "video.frames_per_second": 1,
    }

    def __init__(self, reset_app: Callable[[], WebDriver]) -> None:
        super(WebDriverEnv, self).__init__()
        self._reset_app = reset_app
        self._driver = None
        self._logger = logging.getLogger(__name__)

    def reset(self) -> Union[Dict, np.ndarray]:
        self._driver = self._reset_app()
        self._update_spaces()
        return self._observe()

    def step(self, action: AppAction) -> Tuple[Dict, float, bool, WebDriver]:
        # reward options:
        # - [CURRENT]reward if no exception was rised during the action
        # - len(self.action_space.actions - prev_action_space.actions) - we would like to find more new actions
        reward = 0
        try:
            action.perform()
            reward = 1
        except Exception as e:
            self._logger.error(e)
        finally:
            self._update_spaces()
            obs = self._observe()
            done = len(self.action_space.actions) == 0
            info = self._driver
            return obs, reward, done, info

    def render(self, mode: str = "human") -> Union[Image.Image, str, np.ndarray]:
        pic = self._get_screenshot()

        if mode == "human":
            return pic

        if mode == "base64":
            buffer = BytesIO()
            pic.save(buffer, format="PNG")
            return base64.b64encode(buffer.getvalue()).decode("ascii")

        if mode == "rgb_array":
            dims = pic.size
            return np.array(pic.getdata(), np.uint8).reshape(dims[1], dims[0], 3)

        raise NotImplementedError(f"Mode {mode} is not supported by {__name__}")

    def close(self) -> None:
        try:
            self._driver.close()
            self._driver.quit()
        except:
            pass

    def _observe(self) -> Dict[str, Any]:
        return {"screenshot": self.render(), "has_error": 0}

    def _update_spaces(self) -> None:
        app_window_size = self._driver.get_window_size()
        # do we consider:
        # - window resizes
        # - new windows openings with different size
        self.observation_space = spaces.Dict(
            {
                "screenshot": spaces.Box(
                    low=0,  # brightness level low
                    high=255,  # brightness level high
                    shape=(app_window_size["height"], app_window_size["width"], 3),
                    dtype=np.uint8,
                ),
                "has_error": spaces.Discrete(2),
            }
        )
        self.action_space = DiscreteAppActionSpace(self._driver)

    def _get_screenshot(self) -> Image:
        # to take the entire screenshot https://stackoverflow.com/a/53825388 could be used
        body_el = self._driver.find_element_by_tag_name('body')
        self._driver.set_window_size(body_el.size['height'], body_el.size['width'])

        stream = BytesIO(self._driver.get_screenshot_as_png())
        pic = Image.open(stream).convert("RGB")
        stream.close()
        return pic



from typing import Tuple

from PIL import Image, ImageOps


def image_pad_resize_to(im: Image, new_dims: Tuple[int, int]) -> Image:
    curr_dims = im.size

    new_ratio = new_dims[0] / new_dims[1]
    curr_ratio = curr_dims[0] / curr_dims[1]

    new_w = curr_dims[1] * new_ratio
    delta_w = new_w - curr_dims[0] if new_ratio > curr_ratio else 0

    new_h = curr_dims[0] / new_ratio
    delta_h = new_h - curr_dims[1] if new_ratio < curr_ratio else 0

    padding = (round(delta_w / 2), round(delta_h / 2))
    return ImageOps.expand(im, padding, "black").resize(new_dims, Image.ANTIALIAS)



import argparse
import glob
import json
import os
import logging
import re
import numpy as np
import pandas as pd
import xml.etree.ElementTree as ET

from os.path import basename, dirname
from typing import Tuple, List, Dict, Union
from bs4 import BeautifulSoup
from PIL import Image, ImageDraw
from xmldiff import main as diffmain
from datetime import datetime
from jinja2 import Environment, FileSystemLoader
from change_analyzer.wrappers.sequence_recorder import SequenceRecorder


class SequencesDiff:

    SUPPORTED_DIFFS = ['UpdateAttrib']

    def __init__(self, sequence1_file: str, sequence2_file: str) -> None:
        self._logger = logging.getLogger(__name__)
        self.report_date = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
        self.sequence1_date = basename(dirname(sequence1_file))
        self.sequence2_date = basename(dirname(sequence2_file))
        self.expected_sequence_id = basename(sequence1_file).split('.')[0]
        self.actual_sequence_id = basename(sequence2_file).split('.')[0]
        self.report_folder = os.path.join(dirname(sequence1_file), "comparisons", self.report_date)
        os.makedirs(self.report_folder, exist_ok=True)
        self.df_merged = pd.DataFrame()

        self._update_merged_df(sequence1_file, sequence2_file)
        self._write_to_report(os.path.join(self.report_folder, 'index.html'))

    def _update_merged_df(self, sequence1_file, sequence2_file) -> None:
        """Update merged dataframe using expected and actual dataframes"""
        self.df_merged = pd.merge(
            pd.read_csv(sequence1_file),
            pd.read_csv(sequence2_file),
            left_index=True,
            right_index=True,
            suffixes=('_expected', '_actual')
        )

        # Add the first step into DF, based on current first row of DF
        first_step_action = 'open the Application'
        if 'html' in self.df_merged[f'{SequenceRecorder.COL_PAGE_SOURCE_BEFORE}_actual'][0]:
            first_step_action = 'open the Website URL'
        first_step_data = {
            f'{SequenceRecorder.COL_SEQUENCE_ID}_expected': self.df_merged[f'{SequenceRecorder.COL_SEQUENCE_ID}_expected'][0],
            f'{SequenceRecorder.COL_SEQUENCE_ID}_actual': self.df_merged[f'{SequenceRecorder.COL_SEQUENCE_ID}_actual'][0],
            f'{SequenceRecorder.COL_PAGE_SOURCE_AFTER}_actual': self.df_merged[f'{SequenceRecorder.COL_PAGE_SOURCE_BEFORE}_actual'][0],
            f'{SequenceRecorder.COL_PAGE_SOURCE_AFTER}_expected': self.df_merged[f'{SequenceRecorder.COL_PAGE_SOURCE_BEFORE}_expected'][0],
            f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_actual': self.df_merged[f'{SequenceRecorder.COL_ACTION_IMAGE_BEFORE}_actual'][0],
            f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_expected': self.df_merged[f'{SequenceRecorder.COL_ACTION_IMAGE_BEFORE}_expected'][0],
            f'{SequenceRecorder.COL_ACTION_TO_PERFORM}_expected': first_step_action,
            f'{SequenceRecorder.COL_ACTION_TO_PERFORM}_actual': first_step_action,
        }
        first_step_df = pd.DataFrame([first_step_data])
        self.df_merged = pd.concat([first_step_df, self.df_merged], ignore_index=True)
        self.df_merged.to_csv(os.path.join(self.report_folder, 'df.csv.zip'), index=False)

        # Add Platform
        self.df_merged['Platform'] = np.where(self.df_merged['PageSourceAfter_actual'].str.contains('html') |
                                              self.df_merged['PageSourceBefore_actual'].str.contains('html'),
                                              'web', 'win')

        # Images
        img_col = [col for col in self.df_merged.columns if "image" in col.lower()]
        for col in img_col:
            self.df_merged.fillna({col: f"{np.zeros((1,1,3))}"}, inplace=True)
            self.df_merged[col] = self.df_merged[col].apply(self.json_to_image)
        self.df_merged["ImageVerdict"] = self.df_merged.apply(lambda row: "pass" if np.array_equal(
            row[f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_expected'],
            row[f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_actual']
        ) else "fail", axis=1)

        # XMLs
        xml_cols = [col for col in self.df_merged.columns if "pagesource" in col.lower()]
        for col in xml_cols:
            self.df_merged.fillna({col: '<?xml version="1.0" encoding="UTF-8"?><metadata></metadata>'}, inplace=True)
            self.df_merged[col] = self.df_merged[col].apply(self.encode_xml)

        self.df_merged["PageSource_diff"] = self.df_merged.apply(lambda row: self._extract_diffs_as_dict(
            row[f'{SequenceRecorder.COL_PAGE_SOURCE_AFTER}_expected'],
            row[f'{SequenceRecorder.COL_PAGE_SOURCE_AFTER}_actual'],
            row['Platform']
        ), axis=1)
        self.df_merged["DiffInfo"] = self.df_merged["PageSource_diff"].apply(self._get_diff_info)

        # Image highlighting
        self.df_merged.apply(lambda row: self._draw_boundaries(
            row["DiffInfo"],
            row[f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_actual'],
            row[f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_expected'],
        ), axis=1)
        self.df_merged["actual_screenshot_filename"] = self.df_merged.apply(lambda row: f"actual_screenshot_step{row.name + 1}.png", axis=1)
        self.df_merged["expected_screenshot_filename"] = self.df_merged.apply(lambda row: f"expected_screenshot_step{row.name + 1}.png", axis=1)

        self.df_merged.apply(lambda row: self._save_step_images(
            row[f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_actual'],
            row[f'{SequenceRecorder.COL_ACTION_IMAGE_AFTER}_expected'],
            row["actual_screenshot_filename"],
            row["expected_screenshot_filename"],
        ), axis=1)

    @staticmethod
    def json_to_image(img: str) -> Union[Image.Image, float]:
        try:
            return Image.fromarray(np.array(json.loads(img), dtype=np.uint8))
        except:
            return np.nan

    @staticmethod
    def encode_xml(xml: str) -> Union[bytes, float]:
        encodings = re.findall(r'encoding=\"(.+?)\"', xml)
        if len(encodings) == 0:
            return xml.encode("utf-8")
        return xml.encode(encodings[0])

    def _save_step_images(self, image_actual: Image.Image, image_expected: Image.Image, filename_actual: str, filename_expected: str) -> None:
        """Save step expected and actual images of the SUT"""
        expected_image_filepath = os.path.join(self.report_folder, filename_expected)
        actual_image_filepath = os.path.join(self.report_folder, filename_actual)

        image_expected.save(expected_image_filepath, format="PNG")
        image_actual.save(actual_image_filepath, format="PNG")

    @staticmethod
    def _draw_boundaries(info_list: List[Dict], image_actual_pil: Image.Image, image_expected_pil: Image.Image) -> None:
        if len(info_list) == 0:
            return

        diff_type = {
            'UpdateAttrib': 'orange',
            'DeleteAttrib': 'red',
            'InsertAttrib': 'green'
        }
        for info in info_list:
            actual_element_boundaries = info['actual_element_boundaries']
            expected_element_boundaries = info['expected_element_boundaries']

            # Get outline color, based on diff type
            outline_color = diff_type[info['actual_element']['type']]

            # Draw boundaries
            draw_image_actual = ImageDraw.Draw(image_actual_pil)
            draw_image_expected = ImageDraw.Draw(image_expected_pil)
            draw_image_actual.polygon(actual_element_boundaries, fill=None, outline=outline_color)
            draw_image_expected.polygon(expected_element_boundaries, fill=None, outline=outline_color)

    def _write_to_report(self, report_file) -> None:
        """Write the report file using Jinja template"""
        template_folder = os.path.abspath(os.path.join(__file__, "../templates"))
        template = Environment(loader=FileSystemLoader(template_folder), autoescape=True).get_template("Log_template.html")

        self.expected_images = self.df_merged["expected_screenshot_filename"]
        self.actual_images = self.df_merged["actual_screenshot_filename"]
        verdicts = self.df_merged["ImageVerdict"].tolist()
        comments = ["The actual screenshot is the same as the expected screenshot"
                    if verdict == "pass"
                    else "The actual screenshot is not the same as the expected screenshot"
                    for verdict in verdicts]
        expected_steps = self.df_merged[f'{SequenceRecorder.COL_ACTION_TO_PERFORM}_expected'].tolist()
        actual_steps = self.df_merged[f'{SequenceRecorder.COL_ACTION_TO_PERFORM}_actual'].tolist()

        sequences_data = zip(expected_steps, actual_steps,
                             self.expected_images, self.actual_images,
                             verdicts, comments)

        steps = [f"step{i+1}" for i in range(len(expected_steps))]
        aria_controls = ",".join(steps)

        # Render HTML Template String
        html_template_string = template.render(expected_sequence_id=self.expected_sequence_id,
                                               actual_sequence_id=self.actual_sequence_id,
                                               sequences_data=sequences_data,
                                               report_date=self.report_date,
                                               aria_controls=aria_controls,
                                               sequence1_date=self.sequence1_date,
                                               sequence2_date=self.sequence2_date)

        # Save the new HTML file
        with open(report_file, "w") as f:
            f.write(html_template_string)

    @staticmethod
    def _diff_is_valid(diff: str) -> bool:
        """Check if diff doesn't contain ignored diffs
        Currently we don't check a diff further if it contains the following strings:
            - ProcessId
            - RuntimeId
            - script
        """
        ignored_diffs = ['ProcessId', 'RuntimeId', 'script']
        return not any(ignored_diff in diff for ignored_diff in ignored_diffs)

    @staticmethod
    def _get_attribute_value_based_on_node(page_root: ET.Element, node: str, attribute: str) -> str:
        """Get attribute value for specific node from given page source"""
        formatted_node = "./" + "/".join(node.split("/")[2::])
        elem = page_root.findall(formatted_node)[0]
        return elem.attrib[attribute]

    def _extract_diffs_as_dict(self, expected_xml: str, actual_xml: str, platform: str) -> Dict:
        """Extract diffs as dictionary, using expected and actual approach"""
        diffs_as_dict = {
            'expected': {},
            'actual': {}
        }

        if platform == 'web':
            # We need to preprocess the html to be xml-like
            expected_xml = BeautifulSoup(expected_xml, 'html.parser').prettify()
            actual_xml = BeautifulSoup(actual_xml, 'html.parser').prettify()

        expected_root = ET.ElementTree(ET.fromstring(expected_xml)).getroot()
        actual_root = ET.ElementTree(ET.fromstring(actual_xml)).getroot()

        for diff in diffmain.diff_texts(expected_xml, actual_xml):
            diff_type = str(diff).split("(")[0]
            if self._diff_is_valid(str(diff)) and diff_type in self.SUPPORTED_DIFFS:
                expected_value = self._get_attribute_value_based_on_node(expected_root, diff.node, diff.name)
                if diff.node not in diffs_as_dict['actual'].keys():
                    diffs_as_dict['actual'][diff.node] = {diff.name: diff.value}
                    diffs_as_dict['actual'][diff.node] = {'type': diff_type}
                    diffs_as_dict['expected'][diff.node] = {diff.name: expected_value}
                else:
                    diffs_as_dict['actual'][diff.node][diff.name] = diff.value
                    diffs_as_dict['expected'][diff.node][diff.name] = expected_value

        # Update each node from dict with expected and actual x,y,w,h if they are not already present
        for k, v in diffs_as_dict['actual'].items():
            expected_element = diffs_as_dict['expected'][k]
            actual_element = diffs_as_dict['actual'][k]
            if 'x' not in diffs_as_dict['actual'][k].keys():
                expected_x = self._get_attribute_value_based_on_node(expected_root, k, 'x')
                actual_x = self._get_attribute_value_based_on_node(actual_root, k, 'x')
                expected_element['x'] = expected_x
                actual_element['x'] = actual_x
            if 'y' not in diffs_as_dict['actual'][k].keys():
                expected_y = self._get_attribute_value_based_on_node(expected_root, k, 'y')
                actual_y = self._get_attribute_value_based_on_node(actual_root, k, 'y')
                expected_element['y'] = expected_y
                actual_element['y'] = actual_y
            if 'width' not in diffs_as_dict['actual'][k].keys():
                expected_width = self._get_attribute_value_based_on_node(expected_root, k, 'width')
                actual_width = self._get_attribute_value_based_on_node(actual_root, k, 'width')
                expected_element['width'] = expected_width
                actual_element['width'] = actual_width
            if 'height' not in diffs_as_dict['actual'][k].keys():
                expected_height = self._get_attribute_value_based_on_node(expected_root, k, 'height')
                actual_height = self._get_attribute_value_based_on_node(actual_root, k, 'height')
                expected_element['height'] = expected_height
                actual_element['height'] = actual_height

        return diffs_as_dict

    def _get_diff_info(self, diff_results: Dict) -> List[Dict]:
        if diff_results['expected'] == {} or diff_results['actual'] == {}:
            return []

        step_diff_info = []
        for element in diff_results['actual'].keys():
            expected_element = diff_results['expected'][element]
            actual_element = diff_results['actual'][element]

            expected_x = expected_element['x']
            actual_x = actual_element['x']
            expected_y = expected_element['y']
            actual_y = actual_element['y']
            expected_width = expected_element['width']
            actual_width = actual_element['width']
            expected_height = expected_element['height']
            actual_height = actual_element['height']

            expected_box_coordinates = self._get_box_coordinates(expected_x, expected_y,
                                                                 expected_width, expected_height)
            actual_box_coordinates = self._get_box_coordinates(actual_x, actual_y,
                                                               actual_width, actual_height)
            info = f'Actual ({actual_box_coordinates}) vs expected ({expected_box_coordinates})'

            step_diff_info.append({
                'actual_element': actual_element,
                'expected_element': expected_element,
                'actual_element_boundaries': actual_box_coordinates,
                'expected_element_boundaries': expected_box_coordinates,
                'info': info
            })

        return step_diff_info

    @staticmethod
    def _get_box_coordinates(x: str, y: str, w: str, h: str) -> List[Tuple[int, int]]:
        """Get boundary coordinates based on x, y, width and height"""
        x = int(x)
        y = int(y)
        w = int(w)
        h = int(h)
        return [(x, y), (x+w, y), (x+w, y+h), (x, y+h)]


def find_csv_file_within_folder(folder: str) -> str:
    """"Find the sequence csv file within the associated folder, assuming there is only one csv file present"""
    try:
        return glob.glob(os.path.join(os.getcwd(), "recordings", folder, "*.csv"))[0]
    except:
        pass

def find_last_two_valid_folders() -> Tuple[str, str]:
    """Find the last two folders which have a csv file"""
    folders_found = 0
    csv_file_paths = []
    recordings_folder = os.path.join(os.path.abspath(os.getcwd()), 'recordings')
    os.chdir(recordings_folder)
    folders = list(filter(os.path.isdir, os.listdir(recordings_folder)))
    folders = [os.path.join(recordings_folder, f) for f in folders]  # add path to each file
    folders.sort(key=lambda x: os.path.getmtime(x))
    folders.reverse()

    for folder in folders:
        csv_file_path = find_csv_file_within_folder(folder)
        if csv_file_path:
            folders_found += 1
            csv_file_paths.append(csv_file_path)
        if folders_found == 2:
            break

    return csv_file_paths[0], csv_file_paths[1]

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--sequence1_folder",
        help="recordings folder with the expected sequence",
        required=False,
    )
    parser.add_argument(
        "--sequence2_folder",
        help="recordings folder with the expected sequence",
        required=False,
    )
    args = parser.parse_args()

    sequence1_file = ""
    sequence2_file = ""
    if not args.sequence1_folder and not args.sequence2_folder:
        sequence1_file, sequence2_file = find_last_two_valid_folders()

    if args.sequence1_folder and not args.sequence2_folder:
        sequence1_file = find_csv_file_within_folder(args.sequence1_folder)
        sequence2_file, _ = find_last_two_valid_folders()

    if args.sequence1_folder and args.sequence2_folder:
        sequence1_file = find_csv_file_within_folder(args.sequence1_folder)
        sequence2_file = find_csv_file_within_folder(args.sequence2_folder)

    SequencesDiff(sequence1_file, sequence2_file)


if __name__ == "__main__":
    main()



from selenium.webdriver.remote.webelement import WebElement

from change_analyzer.spaces.actions.app_action import AppAction


class FormAppAction(AppAction):
    def __init__(self, form: WebElement) -> None:
        super(FormAppAction, self).__init__(form)

    def perform(self) -> None:
        # TODO basing on el
        #  * generate input values for fields
        #  * handle checkboxes, dropdowns, radios somehow
        #  * perform submit action (click input[type=submit]? what about standalone apps?)
        pass



"""
Setup harness
"""
from setuptools import setup, find_packages
from distutils.util import convert_path


def _read_long_description():
    with open("README.md") as readme:
        return readme.read()


REQUIRE = [
    "gym",
    "pandas",
    "Pillow",
    "pywinauto",
    "selenium",
    "faker",
    "jinja2",
    "pandas",
    "numpy",
    "matplotlib",
    "xmldiff",
    "beautifulsoup4",
    "lxml"
]
DEV_REQUIRE = [
    "black",
    "jupyterlab",
    "pytest",
    "python-semantic-release",
    "twine",
    "wheel",
    "xenon",
]
NAME = "change-analyzer"
NAME_DASHED = NAME.replace("_", "-")

FILE_NAME = convert_path('change_analyzer/__init__.py')
with open(FILE_NAME) as FILE:
    for LINE in FILE:
        if "__version__" in LINE:
            VERSION_NUMBER = LINE[LINE.find('"')+1:LINE.rfind('"')]

setup(
    name=NAME_DASHED,
    description="Change analyzer",
    long_description=_read_long_description(),
    long_description_content_type="text/markdown",
    author="Matvey Pashkovskiy, Sorin Patrasoiu, Joona Oikarinen",
    author_email="",
    url=f"https://stash.f-secure.com/users/pashma/repos/{NAME}",
    platforms="any",
    version=VERSION_NUMBER,
    packages=find_packages(exclude=[f"{NAME_DASHED}.tests", f"{NAME_DASHED}.tests.*"]),
    entry_points={
        "console_scripts": [
            "ca-run=change_analyzer.main:main",
            "ca-compare=change_analyzer.sequences_diff:main"
        ]
    },
    install_requires=REQUIRE,
    extras_require={"dev": DEV_REQUIRE},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import logging
import gym


class Agent:
    def __init__(self, env: gym.Env) -> None:
        self.env = env
        self.total_reward = 0.0
        self._logger = logging.getLogger(__name__)

    def run(self) -> None:
        raise NotImplementedError()



from typing import Any


class TransparentWrapperMixin:
    def __getattribute__(self, name) -> Any:
        # TODO refactor as this extremely dirty hack will work only if AppEnv is the very previous to SequenceRecorder
        # or all the previous wrappers has this mixin
        if name == "action_space":
            return self.env.action_space
        if name == "observation_space":
            return self.env.observation_space

        return object.__getattribute__(self, name)



from random import randrange

import gym

from change_analyzer.agents.agent import Agent


class RandomAgent(Agent):
    def __init__(self, env: gym.Env, total_steps: int = randrange(10)) -> None:
        super(RandomAgent, self).__init__(env)
        self.total_steps = total_steps

    def run(self) -> None:
        for i in range(self.total_steps):
            # Buttons placed in the center of the screen have to be selected more often. It is possible to set higher
            # probability by calculating how far the element from the center of the screen. Or even top left quadrant
            # should be the focus area.
            try:
                action = self.env.action_space.sample()
                print("Perform action ", str(action))
                obs, reward, done, info = self.env.step(action)
                self.total_reward += reward
            except Exception as e:
                self._logger.info("Action couldn't be performed due to an exception")
                self._logger.info(e)
                done = True
            if done:
                break
        self._logger.info(
            f"Episode done in {self.total_steps} steps, total reward {self.total_reward}"
        )
